# -*- coding: utf-8 -*-
"""
Created on Wed Jan 11 10:22:27 2017

@author: akibdpt
"""
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .tree import tree, DescriptionEvaluatingError
from .symbols import symbols
from .numberLiteral import numberLiteral
from .stringLiteral import stringLiteral
from .tuple import tupleDesc
from ...io.parseReader import parseReader, ParsingError
from ...struct.coordinate import coordinate
from ...struct.vector import vector2D, vector3D


class reference(tree):
    type = 'description-reference'
    def __init__(self, sort, rhs = False):
        self.sort = sort
        self.method = None
        self.subSort = None
        self.subMethod = None
        self.subCond = None
        self.value = None
        self.rhs = rhs
        
    def __eq__(self, other):
        if other.type != 'description-reference':
            return False
        return self.sort == other.sort and self.method == other.method
    
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1

    def compareValue(self, other):
        if other.type == 'description-reference':
            c = self.sort.compareValue(other.sort)
            if c > 0:
                return 1
            if c < 0:
                return -1
            if self.method > other.method:
                return 1
            if self.method < other.method:
                return -1
            return 0
        elif (other.type == 'description-stringExpression' or 
                    other.type == 'description-numericExpression' or 
                    other.type == 'description-tupleExpression'):
            return -other.compareValue(self)
        c = hash(self.__class__) - hash(other.__class__)
        if c > 0:
            return 1
        if c < 0:
            return -1
        return 0

    def __str__(self):
        s = ''
        if self.rhs:
            s += 'rhs.'
        s += self.sort.name + symbols.CONCAT + self.method
        if self.subSort is not None:
            s += ':' + str(self.subSort) + symbols.CONCAT + self.subMethod + '=' 
            if type(self.subCond) == str:# or type(self.subCond) == unicode:
                s += '"'+self.subCond+'"'
            else:
                s += str(self.subCond)
        return s

    def printVal(self):
        s = 'ref:'
        if self.rhs:
            s += 'rhs.'
        s += self.sort.name + symbols.CONCAT + self.method
        if self.subSort is not None:
            s += ':' + str(self.subSort) + symbols.CONCAT + self.subMethod + '=' 
            if type(self.subCond) == str:
                s += '"'+self.subCond+'"'
            elif isinstance(self.subCond,int) or isinstance(self.subCond,float):
                s += str(self.subCond)
            else:
                s += self.subCond.printVal()
        return s
        
    def visitEnter(self, element, assignment):
        if self.subMethod == 'value':
            self.subMethod = 'getValue'
        if 'ind' in element.type:
            if element.sort.name == self.sort.name or element.sort.baseSort().name == self.sort.name:
                if self.subSort is not None and not isinstance(self.subSort,str):
                    if isinstance(self.subCond, int) or isinstance(self.subCond, float) or isinstance(self.subCond, str):
                        cmpVal = self.subCond
                    else:
                        cmpVal = self.subCond.evaluate(assignment, self.sort).value
                    if element.hasAttribute():
                        for i in element.attribute:
                            if reference.checkS(i, self.subSort.name, self.subMethod, cmpVal, assignment):
                                if element.type == 'ind-description' and self.method == 'getValue':
                                    self.value = element.evaluateInd(assignment).getValue()[1:-1]
                                    if self.value[0] == '"':
                                        self.value = self.value[1:-1]
                                else:
                                    met = getattr(element, self.method)
                                    if hasattr(met, '__call__'):
                                        try:
                                            self.value = met(assignment)
                                        except TypeError:
                                            self.value = met()
                                    else:
                                        self.value = met
                                return True
                    else:
                        return False
                else:
                    if self.subSort is not None and self.subSort == 'tag':
                        if element.key is not None and str(element.key.name)==str(self.subCond):
                            pass
                        else:
                            return False
                    if element.type == 'ind-description' and self.method == 'getValue':
                        self.value = element.evaluateInd(assignment).getValue()[1:-1]
                        if self.value[0] == '"':
                            self.value = self.value[1:-1]
                    else:
                        met = getattr(element, self.method)
                        if hasattr(met, '__call__'):
                            try:
                                self.value = met(assignment)
                            except TypeError:
                                self.value = met()
                        else:
                            self.value = met
                    return True
            """
            elif self.subSort is not None and element.sort.baseSort().name == self.subSort.name:
                met = getattr(element, self.subMethod)
                if element.type == 'ind-description' and self.subMethod == 'getValue':
                    val = element.evaluateInd(assignment).getValue()[1:-1]
                    if val[0] == '"':
                        val = val[1:-1]
                elif hasattr(met, '__call__'):
                    try:
                        val = met(assignment)
                    except TypeError:
                        val = met()
                else:
                    val = met
                    
                if isinstance(self.subCond, int) or isinstance(self.subCond, float) or isinstance(self.subCond, str):
                    cmpVal = self.subCond
                else:
                    cmpVal = self.subCond.evaluate(assignment, self.sort).value
                    
                if val == cmpVal:
                    if element.hasAttribute():
                        for i in element.attribute:
                            check = reference.checkS(i, self.sort.name, self.method)
                            if check:
                                self.value = check[0]
                                return True
                    else:
                        return False
                else:
                    return False
            """
            return False
        elif 'form' in element.type:
            if element.type == 'form-metaForm' or element.sort.name == self.sort.name or element.sort.baseSort().name == self.sort.name:
                return True
            return False
        
    @staticmethod
    def checkS(ind, sortName, sortMethod, sortValue = None, assignment = None):
        if 'form' in ind.type:
            if ind.type == 'form-metaForm' or ind.sort.baseSort().name == sortName:
                for i in ind:
                    check = reference.checkS(i, sortName, sortMethod, sortValue, assignment)
                    if check:
                        return check
            return False
        if ind.sort.baseSort().name == sortName:
            if ind.type == 'ind-description' and sortMethod == 'getValue':
                val = ind.evaluateInd(assignment).getValue()[1:-1]
                if val[0] == '"':
                    val = val[1:-1]
                if val == sortValue:
                    return True
                else:
                    return False
            met = getattr(ind, sortMethod)
            val = met()
            if sortValue is not None:
                if val == sortValue:
                    return True
                else:
                    return False
            else:
                return [val]
        if ind.hasAttribute():
            for i in ind.attribute:
                check = reference.checkS(i, sortName, sortMethod, sortValue)
                if check:
                    return check
        return False
        
    def visitLeave(self, element):
        return
        
    def evaluate(self, assignment, sort):
        if self.sort.baseSort().characteristic.type == 'ind-description':
            retrv = assignment.retrieve(self.sort, self.method)
            if retrv is None:
                if self.method == 'value' or self.method == 'getValue':
                    pass
                else:
                    raise DescriptionEvaluatingError('parameter name "'+self.method+'" of sort "'+self.sort.name+'" does not exists in the param set: '+str(assignment.parameters))
            else:
                return retrv
        if assignment.lhs is None and assignment.rhs is None:
            raise ValueError('lhs and rhs are not assigned to the paramSet yet')
        if self.subMethod == 'value':
            self.subMethod = 'getValue'
        if self.method == 'value':
            self.method = 'getValue'
        if self.subSort == 'tag' and self.subMethod == 'getValue':
            if self.subCond in assignment.keys:
                ind = assignment.keys[self.subCond].ind.transform(assignment.transform)
                met = getattr(ind, self.method)
                if hasattr(met, '__call__'):
                    try:
                        self.value = met(assignment)
                    except TypeError:
                        self.value = met()
                else:
                    self.value = met
        elif self.rhs == True:
            assignment.rhs.accept(self, assignment)
        else:
            assignment.lhs.accept(self, assignment)
        if self.value is not None:
            if self.sort.baseSort().characteristic.type == 'ind-description':
                try:
                    self.value = coordinate(self.value)
                except:
                    pass
            return self.constructValue(self.value)
        return numberLiteral(0)
        
    @staticmethod
    def constructValue(value):
        if isinstance(value, coordinate):
            return numberLiteral(value)
        elif isinstance(value, int) or isinstance(value, float):
            return numberLiteral(coordinate(value))
        elif isinstance(value, str):
            return numberLiteral(coordinate(value))
        elif isinstance(value, list):
            tup = tupleDesc(symbols.PAREN_OPEN, symbols.COMMA)
            for i in value:
                tup.tuple.append(reference.constructValue(i))
            return tup
        elif isinstance(value, vector2D):
            tup = tupleDesc(symbols.PAREN_OPEN, symbols.COMMA)
            tup.tuple.append(numberLiteral(value.x))
            tup.tuple.append(numberLiteral(value.y))
            return tup
        elif isinstance(value, vector3D):
            tup = tupleDesc(symbols.PAREN_OPEN, symbols.COMMA)
            tup.tuple.append(numberLiteral(value.x))
            tup.tuple.append(numberLiteral(value.y))
            tup.tuple.append(numberLiteral(value.z))
            return tup
        else:
            raise ValueError('Value '+value+': type not recognized.')
            
    """
    def map(self, params, sort, onto):
        if onto is None:
            raise TypeError('description tree cannot be None')
        self.value = onto.value
        if self.rhs:
            params.rhsDelay(self)
        else:
            params.lhsDelay(self)
        return True
        
    def delayedMatch(self, params, form):
        onto = self.value
        form.accept(self)
        if onto == self.value:
            return True
        return False
    """
        
    @staticmethod
    def parse(desc, reader):
        rhs = False
        if reader.value() == 'rhs':
            rhs = True
            reader.newToken()
            reader.newToken()
        name = reader.value()
        if reader.token() != parseReader.IDENTIFIER or reader.newToken() != symbols.CONCAT:
            raise ParsingError(reader, "expected an identifier followed by a '.'")
        if desc.sort.checkSort(name) == False:
            raise ParsingError(reader, "sort '" + name + "' not recognized")
        result = reference(desc.sort.getSort(name), rhs)
        reader.newToken()
        result.method = reader.value()

        if reader.newToken() == ':':
            name = reader.newValue()
            if reader.token() != parseReader.IDENTIFIER or reader.newToken() != symbols.CONCAT:
                raise ParsingError(reader, "expected an identifier followed by a '.'")
            if name == 'tag':
                result.subSort =  name
            else:
                if desc.sort.checkSort(name) == False:
                    raise ParsingError(reader, "sort '" + name + "' not recognized")
                result.subSort = desc.sort.getSort(name)
            reader.newToken()
            result.subMethod = reader.value()
            
            if reader.newToken() != '=':
                raise ParsingError(reader, "expected a '='")

            if reader.newToken() == parseReader.STRING:
                result.subCond = str(reader.value()[1:-1])
            elif reader.token() == parseReader.NUMBER:
                try:
                    result.subCond = int(reader.value())
                except ValueError:
                    result.subCond = float(reader.value())
            elif (reader.token() == parseReader.MINUS and reader.previewToken() == parseReader.NUMBER):
                reader.newToken()
                try:
                    result.subCond = -int(reader.value())
                except ValueError:
                    result.subCond = -float(reader.value())
            else:
                result.subCond = desc.numericExpression.parse(desc, reader)
                return result
            reader.newToken()
        return result
