# -*- coding: utf-8 -*-
"""
Created on Mon Aug  8 15:44:26 2016

@author: akibdpt
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division
from __future__ import absolute_import
from builtins import super
from builtins import str
"""
from future import standard_library
standard_library.install_aliases()
"""
from .sort import sort
from .parameter.argument import argument

class primitiveSort(sort):
    characteristicList = {}
    categories = {}
    parameter = {}
    defaultMapping = {}
    dimensionality = {}
    type = 'sort-primitive'
    
    def __init__(self, name, characteristic, arg = None, *args, **kwargs):
        if self.checkSort(name) == True: 
            raise ValueError('A sort with name '+name+' already exists')
        super(primitiveSort, self).__init__(name)
        self.characteristic = characteristic
        self.behavior = primitiveSort.categories[characteristic]
        self.parameter = primitiveSort.parameter[characteristic]
        self.description = '['+characteristic.type[4:]+']'
        if self.parameter != argument.argumentType.NONE:
            self.arguments = argument.defineArgument(self.parameter, arg, characteristic)
            if self.arguments is not None:
                self.description += ' ('+arg+')'
                
    @classmethod
    def create(cls, name, characteristic, arg = None):
        if sort.checkSort(name) == True: 
            s = sort.getSort(name)
            if s.type != 'sort-primitive' or characteristic != s.characteristic:
                raise ValueError('A sort with name '+name+' already exists')
            return s
        return primitiveSort(name, characteristic, arg)
        
    """
    registering each individual with its category, parameter, defaultMapping and dimensionality
    e.g. a point is of discrete category, no parameter, similarity mapping and discrete dimensionality
         a line segment is of interval category, no parameter, similarity mapping and linear dimensionality
         a label is of discrete category, no parameter, exact mapping and discrete dimensionality
         a weight is of ordinal category, 'upperbound' parameter, exact mapping and linear dimensionality
         etc...
    this function is called specifically in sortal/__init__.py and register all available individuals for
    rule analysis and detection
    """
    @staticmethod
    def register(characteristic, b, p, m, d):
        primitiveSort.characteristicList[characteristic.type[4:].lower()] = characteristic
        primitiveSort.categories[characteristic] = b
        primitiveSort.parameter[characteristic] = p
        primitiveSort.defaultMapping[characteristic] = m
        primitiveSort.dimensionality[characteristic] = d
      
    """
    get respective behavior(category) from an individual
    """
    def getBehavior(characteristic):
        return primitiveSort.categories[characteristic]
    """
    get respective parameter from an individual
    """
    def getParameter(self):
        return primitiveSort.parameter[self.characteristic]
    """
    get respective mapping from an individual
    """
    def mapping(self):
        return primitiveSort.defaultMapping[self.characteristic]
    """
    get respective dimensionality from an individual
    """
    def dimension(self):
        return primitiveSort.dimensionality[self.characteristic]
        
    """
    checking if this sort contains another sort i.e. the two sorts are the same
    """
    def contains(self, other):
        if other.type == 'sort-primitive' and other.name == self.name:
            return True
        return False
        
    """
    creating a new form (discrete form / ordinal form / interval form) of primitive sort
    """    
    def newForm(self):
        form = self.behavior(self)
        return form

