# -*- coding: utf-8 -*-
"""
Created on Mon Oct  3 15:14:00 2016

@author: akibdpt
"""
from __future__ import division
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import absolute_import
#from math import acos, atan2, pi
from builtins import round
"""
from future import standard_library
standard_library.install_aliases()
"""
from builtins import object
from mpmath import mp, acos, atan2, pi, chop, nstr
#from decimal import getcontext, Decimal

#mpmath gmpy2

"""
vector in two-dimensional cartesian space
same as vector2D but all the values are stored as float instead of symbolic representation
--> not exact representation of real values
this is only used during rule analysis and detection as the processing time required
is much lower for float values compared to symbolic mathematics  
"""
class vector2Dfloat(object):
    """
    the precision for comparing float values: 12
    i.e. comparing float values up to 12 decimal places
         i have tried to put it at higher precision, but it doesnt seem to work
    """
    mpdps = 76
    p = 30
    mp.dps = mpdps
    def __init__(self, x, y, infinity = False, w = 1):

        self.type = 'struct-vector2Dfloat'
        self.w = w
        self.x = x
        self.y = y
        
    """
    getter for x value
    """
    @property
    def x(self):
        if self.w != 0:
            return self.__x / self.w
        else:
            return self.__x
            
    """
    setter for x value
    """
    @x.setter
    def x(self, val):
        if self.w != 0:
            self.__x = val * self.w
        else:
            self.__x = val
            
    """
    getter for y value
    """
    @property
    def y(self):
        if self.w != 0:
            return self.__y / self.w
        else:
            return self.__y
            
    """
    setter for y value
    """
    @y.setter
    def y(self, val):
        if self.w != 0:
            self.__y = val * self.w
        else:
            self.__y = val

    def __str__(self):
        return '('+nstr(self.x,3)+','+nstr(self.y,3)+')'
        
    def __abs__(self):
        if self.sign() >= 0:
            return self
        else:
            return vector2Dfloat(-self.x, -self.y)
        
        
    def isInfinite(self):
        return self.w == 0
        
    def simplify(self):
        return self
                        
    def sign(self):
        if round(self.x,self.p) > 0:
            return 1
        if round(self.x,self.p) < 0:
            return -1
        if round(self.y,self.p) > 0:
            return 1
        if round(self.y,self.p) < 0:
            return -1
        return 0
    
    def printVal(self):
        return '('+nstr(self.x,3)+','+nstr(self.y,3)+')'
    
    def isZero(self):
        return (round(self.x,vector3Dfloat.p) == 0 and
                round(self.y,vector3Dfloat.p) == 0)
        
    def parallel(self,other):
        if other.type != 'struct-vector2Dfloat':
            return False
        if self.isZero() or other.isZero():
            return True 
        if (round(self.x, vector2Dfloat.p) == round(other.x, vector2Dfloat.p) 
                and round(self.y, vector2Dfloat.p) == round(other.y, vector2Dfloat.p)):
            return True 
        return round(self.x * other.y, vector2Dfloat.p) == round(self.y * other.x, vector2Dfloat.p)
        
    def perpendicular(self,other):
        if other.type != 'struct-vector2Dfloat':
            return False
        return self.dotproduct(other) == 0
    
    def __eq__(self, other):
        if other.type != 'struct-vector2Dfloat':
            return False
        return (round(self.x-other.x, vector2Dfloat.p) == 0
                and round(self.y-other.y, vector2Dfloat.p) == 0)
        
    def __ne__(self, other):
        return not self == other

    def compareValue(self, other):
        if self.w == other.w:
            diffx = round(self.x-other.x, vector2Dfloat.p)
            if diffx > 0:
                return 1
            elif diffx < 0:
                return -1
            diffy = round(self.y-other.y, vector2Dfloat.p)
            if diffy > 0:
                return 1
            elif diffy < 0:
                return -1
            return 0
        elif self.w == 0:
            sx = round(self.x, vector2Dfloat.p)
            if sx < 0:
                return -1
            elif sx > 0:
                return 1
            ox = round(other.x, vector2Dfloat.p)
            if ox < 0:
                return 1
            elif ox > 0:
                return -1
            else:
                sy = round(self.y, vector2Dfloat.p)
                if sy < 0:
                    return -1
                elif sy > 0:
                    return 1
                oy = round(other.y, vector2Dfloat.p)
                if oy < 0:
                    return 1
                elif oy > 0:
                    return -1
                else:
                    return 0
        elif other.w == 0:
            ox = round(other.x, vector2Dfloat.p)
            if ox < 0:
                return 1
            elif ox > 0:
                return -1
            sx = round(self.x, vector2Dfloat.p)
            if sx < 0:
                return -1
            elif sx > 0:
                return 1
            else:
                oy = round(other.y, vector2Dfloat.p)
                if oy < 0:
                    return 1
                elif oy > 0:
                    return -1
                sy = round(self.y, vector2Dfloat.p)
                if sy < 0:
                    return -1
                elif sy > 0:
                    return 1
                else:
                    return 0
                    
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __le__(self, other):
        return self.compareValue(other) < 1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __ge__(self, other):
        return self.compareValue(other) > -1
    
    def scale(self, other):
        if not isinstance(other, vector2Dfloat):
            try:
                other = other.convertFloat()
            except:
                return vector2Dfloat(self.x*other, self.y*other)
        scalar = (other.dotproduct(self)) / (self.dotproduct(self))
        return vector2Dfloat(self.x*scalar, self.y*scalar)
            
    def abs(self):
        return self.scale(self.sign())
                
    def inverse(self):
        return vector2Dfloat(-self.x, -self.y)
        
    def length(self):
        a = (self.x * self.x + self.y * self.y) ** mp.mpf('0.5')
        return a
        
    def squareLength(self):
        return self.x * self.x + self.y * self.y
        
    def unit(self):
        a = 1 / self.length()
        return vector2Dfloat(self.x * a, self.y * a)
    
    def add(self,other):
        if self.w == 0:
            if other.w == 0:
                raise ValueError('Adding two infinite vectors is undefined')
            else:
                return self
        if other.w == 0:
            return other
        return vector2Dfloat(self.x+other.x, self.y+other.y)
    
    def subtract(self, other):
        if self.w == 0:
            if other.w == 0 and self != other.inverse():
                return vector2Dfloat(self.x-other.x, self.y-other.y, True)
            else:
                return self
        if other.w == 0:
            return other
        
        return vector2Dfloat(self.x-other.x, self.y-other.y)

    def dotproduct(self, other):
        return self.x * other.x + self.y * other.y

    def product(self, other):
        return self.x * other.y - self.y * other.x
        
    def scalar(self, other):
        return other.dotproduct(other) / self.dotproduct(self)
    
    def colinear(self, other1, other2):
        return other1.subtract(self).parallel(other2.subtract(self))
    
    def angle(self, other):
        a = self.unit()
        b = other.unit()
        #return (atan2(b.y, b.x) - atan2(a.y, a.x))
        
        c = (atan2(b.y, b.x) - atan2(a.y, a.x))
        
        
        if (c < 0 and c > -pi) or (c > pi and c < 2*pi):            
            return - acos(self.unit().dotproduct(other.unit()))
        return acos(self.unit().dotproduct(other.unit()))
                
    def transform(self, tf):
        
        x = self.x*tf.matrix[0][0] + self.y*tf.matrix[0][1] + tf.matrix[0][2]
        y = self.x*tf.matrix[1][0] + self.y*tf.matrix[1][1] + tf.matrix[1][2]
        return vector2Dfloat(chop(x), chop(y))
                        
    def convertFloat(self):
        return self


class vector3Dfloat(object):
    """
    the precision for comparing float values: 12
    i.e. comparing float values up to 12 decimal places
         i have tried to put it at higher precision, but it doesnt seem to work
    """
    p = vector2Dfloat.p
    
    def __init__(self, x, y, z, infinity = False, w = 1):
        self.type = 'struct-vector3Dfloat'
        self.w = w
        self.x = x
        self.y = y
        self.z = z
        
    """
    getter for x value
    """
    @property
    def x(self):
        if self.w != 0:
            return self.__x / self.w
        else:
            return self.__x
            
    """
    setter for x value
    """
    @x.setter
    def x(self, val):
        if self.w != 0:
            self.__x = val * self.w
        else:
            self.__x = val
            
    """
    getter for y value
    """
    @property
    def y(self):
        if self.w != 0:
            return self.__y / self.w
        else:
            return self.__y
            
    """
    setter for y value
    """
    @y.setter
    def y(self, val):
        if self.w != 0:
            self.__y = val * self.w
        else:
            self.__y = val
            
    @property
    def z(self):
        if self.w != 0:
            return self.__z / self.w
        else:
            return self.__z
        
    @z.setter
    def z(self, val):
        if self.w != 0:
            self.__z = val * self.w
        else:
            self.__z = val

    def __str__(self):
        if self.w == 0:
            return ' inf('+nstr(self.x,3)+','+nstr(self.y,3)+','+nstr(self.z,3)+')'
        return '('+nstr(self.x,3)+','+nstr(self.y,3)+','+nstr(self.z,3)+')'
        
    def __abs__(self):
        if self.sign() >= 0:
            return self
        else:
            return vector3Dfloat(-self.x, -self.y, -self.z)

    def isInfinite(self):
        return self.w == 0
        
    def simplify(self):
        return self
                        
    def sign(self):
        if round(self.x,self.p) > 0:
            return 1
        if round(self.x,self.p) < 0:
            return -1
        if round(self.y,self.p) > 0:
            return 1
        if round(self.y,self.p) < 0:
            return -1
        if round(self.z,self.p) > 0:
            return 1
        if round(self.z,self.p) < 0:
            return -1
        return 0
    
    def printVal(self):
        return '('+nstr(self.x,3)+','+nstr(self.y,3)+','+nstr(self.z,3)+')'
    
    def isZero(self):
        return (round(self.x,vector3Dfloat.p) == 0 and
                round(self.y,vector3Dfloat.p) == 0 and
                round(self.z,vector3Dfloat.p) == 0)
        
    def parallel(self,other):
        if other.type != 'struct-vector3Dfloat':
            return False
        if self.isZero() or other.isZero():
            return True 
        if (round(self.x, vector3Dfloat.p) == round(other.x, vector3Dfloat.p) and 
            round(self.y, vector3Dfloat.p) == round(other.y, vector3Dfloat.p) and 
            round(self.z, vector3Dfloat.p) == round(other.z, vector3Dfloat.p)):
            return True 
        if round(self.x, vector3Dfloat.p) != 0 or round(other.x, vector3Dfloat.p) != 0:
            return (round(self.x * other.y, vector3Dfloat.p) == round(self.y * other.x, vector3Dfloat.p) and
                    round(self.x * other.z, vector3Dfloat.p) == round(self.z * other.x, vector3Dfloat.p))
        return round(self.y * other.z, vector3Dfloat.p) == round(self.z * other.y, vector3Dfloat.p)
        
    def perpendicular(self,other):
        if other.type != 'struct-vector3Dfloat':
            return False
        return self.dotproduct(other) == 0
    
    def __eq__(self, other):
        if other.type != 'struct-vector3Dfloat':
            return False
        return (round(self.x-other.x, vector3Dfloat.p) == 0 and 
                round(self.y-other.y, vector3Dfloat.p) == 0 and
                round(self.z-other.z, vector3Dfloat.p) == 0)
        
    def __ne__(self, other):
        return not self == other
 

    def compareValue(self, other):
        if self.w == other.w:
            diffx = round(self.x-other.x, vector3Dfloat.p)
            if diffx > 0:
                return 1
            elif diffx < 0:
                return -1
            diffy = round(self.y-other.y, vector3Dfloat.p)
            if diffy > 0:
                return 1
            elif diffy < 0:
                return -1
            diffz = round(self.z-other.z, vector3Dfloat.p)
            if diffz > 0:
                return 1
            elif diffz < 0:
                return -1
            return 0
        elif self.w == 0:
            sx = round(self.x, vector3Dfloat.p)
            if sx < 0:
                return -1
            elif sx > 0:
                return 1
            ox = round(other.x, vector3Dfloat.p)
            if ox < 0:
                return 1
            elif ox > 0:
                return -1
            else:
                sy = round(self.y, vector3Dfloat.p)
                if sy < 0:
                    return -1
                elif sy > 0:
                    return 1
                oy = round(other.y, vector3Dfloat.p)
                if oy < 0:
                    return 1
                elif oy > 0:
                    return -1
                else:
                    sz = round(self.z, vector3Dfloat.p)
                    if sz < 0:
                        return -1
                    elif sz > 0:
                        return 1
                    oz = round(other.z, vector3Dfloat.p)
                    if oz < 0:
                        return 1
                    elif oz > 0:
                        return -1
                    else:
                        return 0
        elif other.w == 0:
            ox = round(other.x, vector3Dfloat.p)
            if ox < 0:
                return 1
            elif ox > 0:
                return -1
            sx = round(self.x, vector3Dfloat.p)
            if sx < 0:
                return -1
            elif sx > 0:
                return 1
            else:
                oy = round(other.y, vector3Dfloat.p)
                if oy < 0:
                    return 1
                elif oy > 0:
                    return -1
                sy = round(self.y, vector3Dfloat.p)
                if sy < 0:
                    return -1
                elif sy > 0:
                    return 1
                else:
                    oz = round(other.z, vector3Dfloat.p)
                    if oz < 0:
                        return 1
                    elif oz > 0:
                        return -1
                    sz = round(self.z, vector3Dfloat.p)
                    if sz < 0:
                        return -1
                    elif sz > 0:
                        return 1
                    else:
                        return 0
                    
    def __lt__(self, other):
        return self.compareValue(other) == -1
        
    def __le__(self, other):
        return self.compareValue(other) < 1
        
    def __gt__(self, other):
        return self.compareValue(other) == 1
        
    def __ge__(self, other):
        return self.compareValue(other) > -1
    
    def scale(self, other):
        if not isinstance(other, vector3Dfloat):
            try:
                other = other.convertFloat()
            except:
                return vector3Dfloat(self.x*other, self.y*other, self.z*other)
        scalar = (other.dotproduct(self)) / (self.dotproduct(self))
        return vector3Dfloat(self.x*scalar, self.y*scalar, self.z*scalar)
            
    def abs(self):
        return self.scale(self.sign())
        
    def normalize(self):
        a = self.length()
        return vector3Dfloat(self.x / a, self.y / a, self.y / a)
        
    def inverse(self):
        return vector3Dfloat(-self.x, -self.y, -self.z)
        
    def length(self):
        a = (self.x * self.x + self.y * self.y + self.z * self.z) ** mp.mpf('0.5')
        return a
        
    def squareLength(self):
        return self.x * self.x + self.y * self.y + self.z * self.z
        
    def unit(self):
        a = 1 / self.length()
        return vector3Dfloat(self.x * a, self.y * a, self.z * a)
    
    def add(self,other):
        if self.w == 0:
            if other.w == 0:
                raise ValueError('Adding two infinite vectors is undefined')
            else:
                return self
        if other.w == 0:
            return other
        return vector3Dfloat(self.x+other.x, self.y+other.y, self.z+other.z)
    
    def subtract(self, other):
        if self.w == 0:
            if other.w == 0 and self != other.inverse():
                return vector3Dfloat(self.x-other.x, self.y-other.y, self.z-other.z, True)
            else:
                return self
        if other.w == 0:
            return other
        return vector3Dfloat(self.x-other.x, self.y-other.y, self.z-other.z)

    def dotproduct(self, other):
        return self.x * other.x + self.y * other.y + self.z * other.z

    def product(self, other):
        return vector3Dfloat(self.y * other.z - self.z * other.y,
                             self.z * other.x - self.x * other.z,
                             self.x * other.y - self.y * other.x)
        
    def scalar(self, other):
        return other.dotproduct(other) / self.dotproduct(self)
    
    def colinear(self, other1, other2):
        return other1.subtract(self).parallel(other2.subtract(self))
    
    """
    def angle(self, other):
        a = self.unit()
        b = other.unit()
        #return (atan2(b.y, b.x) - atan2(a.y, a.x))
        
        c = (atan2(b.y, b.x) - atan2(a.y, a.x))
        
        if (c < 0 and c > -pi) or (c > pi and c < 2*pi):            
            return - acos(self.unit().dotproduct(other.unit()))
        return acos(self.unit().dotproduct(other.unit()))
    """
                
    def transform(self, transf):
        x = (self.x*transf.matrix[0][0] + self.y*transf.matrix[0][1] + self.z*transf.matrix[0][2] + transf.matrix[0][3])
        y = (self.x*transf.matrix[1][0] + self.y*transf.matrix[1][1] + self.z*transf.matrix[1][2] + transf.matrix[1][3])
        z = (self.x*transf.matrix[2][0] + self.y*transf.matrix[2][1] + self.z*transf.matrix[2][2] + transf.matrix[2][3])
        return vector3Dfloat(chop(x), chop(y), chop(z))
                        
    def convertFloat(self):
        return self
